from .utils import read_var
from .auth import _get_token
import requests, json
import datetime
import time


def start(bda_id):
    url = read_var("url_base") + read_var("url_apps") + "/" + bda_id + "/start/" + str(int(bda_id)+1)
    
    payload = json.dumps({})
    headers = {
    'authorization': 'Bearer ' + _get_token(),
    'content-type': 'application/json'
    }

    return requests.request("POST", url, headers=headers, data=payload)

def get_info(bda_id):

    url = read_var('URL_BASE') + read_var('URL_APPS') +"/" + bda_id

    payload = json.dumps({})
    headers = {
    'authorization': 'Bearer ' + _get_token(),
    'content-type': 'application/json'
    }

    response = requests.request("GET", url, headers=headers, data=payload)

    return json.loads(response.text)

def latest_run_status(bda_id):
    info = get_info(bda_id)
    
    dateOfLastStatus = datetime.datetime.strptime("2000-01-01T00:00:00", '%Y-%m-%dT%H:%M:%S')

    for element in info['workflows']:
        for run in element['runs']:
            curr_date = datetime.datetime.strptime(run['dateOfLastStatus'][:-10], '%Y-%m-%dT%H:%M:%S')
            if curr_date>dateOfLastStatus:
                dateOfLastStatus = curr_date
                id = run['k8sRunId']
                
    for element in info['workflows']:
        for run in element['runs']:
            if run['k8sRunId'] == id:
                return run['currentStatus']

def synchronous_wait_till_ends(bda_id, timeout = 60):
    time.sleep(4) # Remove this with a better way of checking that the app is running first
    start = time.time()
    # Wait while BDA is running or until the defined timeout
    while latest_run_status(bda_id=bda_id) == "RUNNING":
        time.sleep(10)
        print("App is running, wait...")
        if time.time() - start > int(timeout):
            break

def start_with_parameters(params, bda_id, wf_id = None):

    if wf_id is None:
        wf_id = str(int(bda_id)+1)

    url = read_var('URL_BASE') + read_var('URL_APPS') + "/" + bda_id + "/start/" + wf_id
    print(bda_id, wf_id)
    print(url)
    payload = json.dumps({
        "services": params
    })
    
    headers = {
        'authorization': 'Bearer ' + _get_token(),
        'content-type': 'application/json',
    }

    response = requests.request("POST", url, headers=headers, data=payload)
        
    return response 


def get_all(order="id"):

    payload = {}
    headers = {
        'authorization': 'Bearer ' + _get_token(),
    }

    response = requests.request("GET", url = read_var('URL_BASE') + read_var('URL_APPS') + "?order="+order, headers=headers, data=payload)

    return response

def get_by_name(name):
    bdas = []
    response = get_all()

    for element in json.loads(response.text)['collection']:
        if element['name'].lower()==name.lower():
            bdas.append(element)

    return bdas, response

def create(bda):
    payload = bda
    headers = {
        'authorization': 'Bearer ' + _get_token(),
        'content-type': 'application/json'
    }
    response = requests.request("POST", url = read_var('URL_BASE') + read_var('URL_APPS'), headers=headers, data=json.dumps(payload))
    return response
    


import requests
import json
from .utils import read_var, update_config_property


def _get_token():
    if read_var("token") is not None:
        return read_var("token")
    else:
        if read_var("username") is not None and read_var("password") is not None: 
            token, _ = get_token(read_var("username"), read_var("password"))
            return token
        else:
            print("WARKING: no token found, please set user and password ENV variables!")

def get_token(user, password):

    payload = 'client_secret=' + read_var('INSTANCE_CLIENT_SECRET') + '&client_id=' + read_var('INSTANCE_CLIENT_ID') + '&grant_type=password&scope=openid&username='+ user +'&password=' + password
    headers = {
    'Content-Type': 'application/x-www-form-urlencoded'
    }

    response = requests.request("POST", url=read_var("url_login"), headers=headers, data=payload)

    try:
        token = json.loads(response.text)['access_token']
    except:
        token = None
    
    update_config_property(prop="token", value=token)
    
    return token, response



from .utils import read_var
from .auth import _get_token
import requests, json
from .minio_utils import download_folder, upload_folder
import os, shutil
import uuid
import pandas as pd
import alidaapis.datasource as ds

def get_metadata(dataset_id):

    url = read_var("URL_BASE") + read_var("URL_DATASETS") + "/" + str(dataset_id)

    payload = {}
    headers = {
        'authorization': 'Bearer ' + _get_token(),
    }

    return requests.request("GET", url, headers=headers, data=payload)


# Starting from a CSV file, infer metadata
def get_cols_metadata_from_csv(path):
    df = pd.read_csv(path)

    pd_types_to_platform_types = {
        "int64": "Number",
        "float64": "Number",
        "object": "String",
        "bool": "String",
        "datatime": "Date"
    }

    col_types = dict(df.dtypes)

    columns = []
    for col in col_types.keys():
        columns.append({"name": col, "type": pd_types_to_platform_types[str(col_types[col])]})

    return columns


def register(remote_path, name, description=None, datasource_id=None, datasource_name=None, tags=[]):
    """Register a dataset composed of generic files and folders (not a tabular dataset)

    Parameters
    ----------
    remote_path : str
        Path inside the datasource
    name :str
        Name of the dataset
    description: str, optional
        Description of the dataset
    datasource_id: str
        The id of the datasource. If None datasource_name will be used
    datasource_name: str, optional
        The name of the datasource. 
    tags: list<str>, optional
        List of tags.
    Returns
    -------
    Http response
    """
    
    if datasource_id is None:
        datasource = ds.get_by_name(datasource_name)
    elif datasource_name is None: 
        datasource = ds.get_by_id(datasource_id)
    else:
        raise Exception("Please specify either datasource id or datasource name")

    
    url = read_var("URL_BASE") + read_var("URL_DATASETS")
    
    payload = {
        "datasetFileType": {
            "type": "image",
            "path": remote_path
        },
        "datasource": datasource,
        "name": name,
        "description": description,
        "tags": tags
    }
    headers = {
        'authorization': 'Bearer ' + _get_token(),
        'content-type': 'application/json'
    }
    return requests.request("POST", url, headers=headers, data=json.dumps(payload))


    
def register_tabular(remote_path, name, path=None, description=None, columns_metadata=None, datasource_id=None, datasource_name=None, tags=[]):
    if columns_metadata is None:
        columns_metadata = get_cols_metadata_from_csv(path=path)
    
    if datasource_id is None:
        datasource = ds.get_by_name(datasource_name)
    else: 
        datasource = ds.get_by_id(datasource_id)
    
    url = read_var("URL_BASE") + read_var("URL_DATASETS")

    payload = {
        "datasetFileType": {
            "type": "table",
            "columns": columns_metadata,
            "tableName": remote_path,
            "path": None
        },
        "datasource": datasource,
        "name": name,
        "description": description,
        "tags": tags
    }
    headers = {
        'authorization': 'Bearer ' + _get_token(),
        'content-type': 'application/json'
    }
    return requests.request("POST", url, headers=headers, data=json.dumps(payload))



def upload_and_register_tabular_from_file(path, name, description=None, datasource_id=None, datasource_name=None, tags=[]):
    if datasource_id is None:
        datasource = ds.get_by_name(datasource_name)
    else: 
        datasource = ds.get_by_id(datasource_id)

    # create tmp folder
    folder_id = str(uuid.uuid4())
    os.mkdir(folder_id)
    shutil.copy(src=path, dst = folder_id + "/")
    remote_path = os.path.join(datasource['prefixPath'], folder_id)
    
    endpoint_url = datasource['host'] + ":" + str(datasource['port'])

    # Check if ssl or not
    if not endpoint_url.startswith('http'):
        if datasource['secure']== True:
            endpoint_url = "https://" + endpoint_url
        else:
            endpoint_url = "http://" + endpoint_url
    
    # Upload dataset to Minio
    for result in upload_folder(minio_address=endpoint_url, 
                                minio_access_key=datasource['accessKey'], 
                                minio_secret_key=datasource['secretKey'], 
                                bucket_name=datasource['bucket'], 
                                local_path=folder_id, 
                                remote_path=remote_path):
        pass
    
    # Remove tmp folder
    shutil.rmtree(folder_id)
    
    response = register_tabular(path=path, remote_path=remote_path, name=name, description=description, datasource_id=datasource_id, datasource_name=datasource_name)
    return response



def download(dataset_id, local_path="./tmp/data/"):
    
    metadata = json.loads(get_metadata(dataset_id=dataset_id).text)
    datasource = metadata['datasource']
    access_key = datasource['accessKey']
    secret_key = datasource['secretKey']
    bucket_name = datasource['bucket']
    endpoint_url = datasource['host'] + ":" + str(datasource['port'])
    if not endpoint_url.startswith('http'):
        if datasource['secure']== True:
            endpoint_url = "https://" + endpoint_url
        else:
            endpoint_url = "http://" + endpoint_url
    if 'path' in metadata['datasetFileType'] and metadata['datasetFileType']['path'] is not None:
        path = metadata['datasetFileType']['path']
    else:
        path = metadata['datasetFileType']['tableName']

    if path[-1]!="/":
        path+="/"
            
    for result in download_folder(minio_address=endpoint_url, minio_access_key=access_key, minio_secret_key=secret_key, bucket_name=bucket_name, local_path=local_path, remote_path=path):
        pass



import os
from os import listdir

def read_var(name):
    return os.environ.get(name.upper())

def update_config_property(prop, value):
    os.environ[prop.upper()] = value
    

def find_csv_filenames( path_to_dir, suffix=".csv" ):
    filenames = listdir(path_to_dir)
    return [ filename for filename in filenames if filename.endswith( suffix ) ]



from .utils import read_var
from .auth import _get_token
import requests, json

def get_all():
    
    url = read_var("url_base") + read_var("url_datasources")
    
    payload = {}
    headers = {
        'authorization': 'Bearer ' + _get_token(),
    }

    return requests.request("GET", url, headers=headers, data=payload)

def get_by_id(id):
    datasources = json.loads(get_all().text)
    for datasource in datasources:
        if int(datasource['id']) == int(id):
            return datasource


def get_by_name(name):
    datasources = json.loads(get_all().text)
    for datasource in datasources:
        if datasource['name'] == name:
            return datasource



import requests
import json
from .auth import _get_token
from .utils import read_var


def get(order = "id", desc="true", limit=20, page=1):
    url = read_var("url_base") + read_var("url_services") + "?sort=" + order + "&desc=" + desc + "&limit=" + str(limit) + "&page=" + str(page)

    payload = {}
    headers = {
        'authorization': 'Bearer ' + _get_token()
    }
    return json.loads(requests.request("GET", url, headers=headers, data=payload).text)

def get_services_by_name(name):
    result = []
    services = json.loads(get().text)['collection']
    
    for service in services:
        if service['name'] == name:
            result.append(service)
    return result

def get_by_id(id):
    url = read_var("URL_ADD_SERVICES") + "/" + str(id)
    payload = {}
    headers = {
        'authorization': 'Bearer ' + _get_token()
    }
    return json.loads(requests.request("GET", url, headers=headers, data=payload).text)


def add(service_metamodel, auto_update_version=False):
    
    if auto_update_version:
        name = service_metamodel['name']
        version = service_metamodel['version']
        if len(get_services_by_name(name))>0:
            latest = get_services_by_name(name)[0]
            vers = latest['version'].split(".")
            vers[-1] = str(int(vers[-1]) + 1)
            version = ".".join(vers)
            service_metamodel['version'] = version
            print(latest)
            print(version)

    
    url = read_var("url_add_services")

    payload = service_metamodel
    headers = {
        'authorization': 'Bearer ' + _get_token(),
        'content-type': 'application/json'
    }
    return requests.request("POST", url, headers=headers, data=json.dumps(payload))



import setuptools

setuptools.setup(
    name="alida-apis",
    version="0.0.8",
    author="Alida research team",
    author_email="salvatore.cipolla@eng.it",
    description="Python APIs for interaction with ALIDA backend.",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = [
        "requests",
        "boto3",
        "minio",
        "progressbar",
        "pandas"
        ],
)



import os
import boto3
import progressbar


def upload_file(bucket, file_name, remote_path):
    print("Uploading: ", file_name)
    
    statinfo = os.stat(file_name)

    up_progress = progressbar.progressbar.ProgressBar(maxval=statinfo.st_size)
    #tmp['max'] = statinfo.st_size
    
    up_progress.start()

    def upload_progress(chunk):
        #tmp['current'] = up_progress.currval + chunk
        #print("Mine: ", int(tmp['current']/tmp['max']*100), "   ")
        # tmp['prog'].update(30)
        up_progress.update(up_progress.currval + chunk)

    bucket.upload_file(file_name, remote_path.replace("\\", "/"), Callback=upload_progress)

    up_progress.finish()

def upload_s3(bucket, local_path, remote_path):
    if os.path.isfile(local_path):  
        bucket.upload_file(local_path, "/".join(remote_path.split("/")[1:]))
        print(local_path)
    else:
        try:
            for path, subdirs, files in os.walk(local_path):
                for file in files:
                    dest_path = path.replace(local_path,"")
                    __s3file = os.path.normpath(remote_path + "/" + dest_path + "/" + file)
                    __local_file = os.path.join(path, file)
                    
                    upload_file(bucket, file_name=__local_file, remote_path=__s3file)
                    yield 1
        except Exception as e:
            print(" ... Failed!! Quitting upload!!")
            print(e)
            raise e



def count_files(path):
    count = 0
    for base, dirs, files in os.walk(path):
        count = count + len(files)    
    return count



def count_files_minio(bucket, prefix):
    count=0
    for obj in bucket.objects.filter(Prefix=prefix):
        count+=1
    return count


def upload_folder(minio_address, minio_access_key, minio_secret_key, bucket_name, local_path, remote_path):

    s3 = boto3.resource('s3', 
        endpoint_url = minio_address,
        aws_access_key_id = minio_access_key, 
        aws_secret_access_key = minio_secret_key
        )
    
    n_of_files = count_files(local_path)

    bucket = s3.Bucket(bucket_name)

    count = 0
    for result in upload_s3(bucket, local_path, remote_path):
        count += result
        yield count, n_of_files


def folder_exists(minio_address, minio_access_key, minio_secret_key, bucket_name, path):
    s3 = boto3.resource('s3', 
        endpoint_url = minio_address,
        aws_access_key_id = minio_access_key, 
        aws_secret_access_key = minio_secret_key
        )
    bucket = s3.Bucket(bucket_name)
    
    if sum(1 for _ in bucket.objects.filter(Prefix=path)) > 0:
        return True
    else:
        return False

def minio_connection_is_working(minio_address, minio_access_key, minio_secret_key, bucket_name):
    s3 = boto3.resource('s3', 
        endpoint_url = minio_address,
        aws_access_key_id = minio_access_key, 
        aws_secret_access_key = minio_secret_key
        )
    bucket = s3.Bucket(bucket_name)

    if bucket.creation_date:
        return True
    else:
        return False

def isfile_s3(bucket, key: str) -> bool:
    """Returns T/F whether the file exists."""
    objs = list(bucket.objects.filter(Prefix=key))
    return len(objs) == 1 and objs[0].key == key


def download_s3(minio_address, minio_access_key, minio_secret_key, bucket_name, remote_path, local_path):
    s3 = boto3.resource('s3', 
        endpoint_url = minio_address,
        aws_access_key_id = minio_access_key, 
        aws_secret_access_key = minio_secret_key
        )
    bucket = s3.Bucket(bucket_name)

    if isfile_s3(bucket, remote_path):
        bucket.download_file(remote_path, local_path)
    else:
        try:
            for obj in bucket.objects.filter(Prefix=remote_path):
                target = obj.key if local_path is None \
                    else os.path.join(local_path, os.path.relpath(obj.key, remote_path))
                if not os.path.exists(os.path.dirname(target)):
                    os.makedirs(os.path.dirname(target))
                if obj.key[-1] == '/':
                    continue
                bucket.download_file(obj.key, target)
                yield 1
        except Exception as e:
            print(" ... Failed!! Quitting download!!")
            print(e)
            raise e


def download_folder(minio_address, minio_access_key, minio_secret_key, bucket_name, local_path, remote_path):

    s3 = boto3.resource('s3', 
            endpoint_url = minio_address,
            aws_access_key_id = minio_access_key, 
            aws_secret_access_key = minio_secret_key
            )
    bucket = s3.Bucket(bucket_name)

    n_of_files = count_files_minio(bucket, remote_path)

    count = 0
    for result in download_s3(bucket_name=bucket_name, local_path=local_path, minio_access_key=minio_access_key, minio_secret_key=minio_secret_key, minio_address=minio_address, remote_path=remote_path):
        count += result
        yield count, n_of_files
